/*
 *
 * Joel's Customized Daemon Kit for Linux
 *
 * From : Unix Network Programming
 *
 */

#include <stdio.h>
#include <signal.h>
#include <sys/param.h>
#include <errno.h>
#include <sys/wait.h>
#include <unistd.h>
#include <sys/stat.h>

#include "daemon.h"

extern int errno;

#ifdef SIGTSP
#	include <sys/file.h>
#	include <sys/ioctl.h>
#endif

/*
 * Detach a daemon process from login session context.
 *
 * if ignsigcld is true, then handle SIGCLDs and keep zombies away.
 */
void daemon_start(int ignsigcld)
{
	register int childpid, fd;

	/*
	 * If we were started by init (process 1) from the /etc/inittab file,
	 * there is no need to detach.
	 * This test is unreliable due to an unavoidable ambiguity if the process
	 * is started by some other process and orphaned (i.e., if the parent process
	 * terminates before we are started).
	 */

	if (getppid() == 1)
		goto out;

	/*
	 * Ignore terminal stop signals (BSD).
	 * Also, deal with hybrids like Linux
	 */

#ifdef SIGTTOU
	signal(SIGTTOU, SIG_IGN);
#endif
#ifdef SIGTTIN
	signal(SIGTTIN, SIG_IGN);
#endif
#ifdef SIGTSTP
	signal(SIGTSTP, SIG_IGN);
#endif

	/*
	 * Make sure we get into the background.  If we are in the foreground,
	 * we had better fork and let the parent exit.  This also makes sure
	 * that we are not a process group leader.
	 */

	if ((childpid = fork()) < 0)
		perror("can't fork first child");
	else if (childpid > 0)
		exit(0);	/* Parent */
	
	/*
	 * First chidl process.
	 *
	 * Disassociate from controlling terminal and process group.
	 * Ensure the process can't reacquire a new controlling terminal.
	 */

#ifdef SIGTSP	/* BSD/LINUX */
	if (setpgrop(0, getpid()) == -1)
		perror("can't change process group");
	
	if ((fd = open("/dev/tty", O_RDWR)) >= 0) {
		ioctl(fd, TIOCNOTTY, (char *) NULL); /* Lose controlling TTY */
		close(fd);
	}

#else /* System V, A/UX, and other brain-dead systems */
	if (setpgrp() == -1)
		perror("can't change process group");
	
	signal(SIGHUP, SIG_IGN); /* Immune from pgrp leader death */

	if ((childpid = fork()) < 0)
		perror("can't fork second child");
	else if (childpid > 0)
		exit(0); /* First child */
	
	/* Second child */

#endif

out:
	/*
	 * Close any open file descriptors
	 */

	for (fd = 0; fd < NOFILE; fd++)
		close(fd);
	
	/* stdin, stdout, and stderr are no longer attached! */

	errno = 0; /* Probably set to EBADF from close */

	/*
	 * Make sure we don't sit on a mounted filesystem
	 */

	chdir("/");

	/*
	 * Clear any inherited file mode creation mask
	 */

	umask(0);

	/*
	 * See if caller isn't interested in exist status of its children
	 * and doesn't want them to become zombies and clog up the system.
	 * With System V, we just ingnore the signal.
	 * With BSD, however, we have to catch each signal and do a wait3.
	 */

	if (ignsigcld) {
#ifdef SIGTSTP
		signal(SIGCLD, sig_child);
#else
		signal(SIGCLD, SIG_IGN);
#endif
	}
}

void sig_child(int sig)
{
#ifdef SIGTSTP
	/*
	 * Use wait3() with WNOHANG
	 */

	int pid;
	union wait status;

	sig = 0;

	while ((pid = wait3(&status, WNOHANG, (struct rusage *) 0)) > 0);
#endif
}
